# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#           Sayan Chowdhury <sayanchowdhury@fedoraproject.org>

import toml

# Read in config file
with open("/etc/fedimg/fedimg-conf.toml") as conffile:
    config = toml.loads(conffile.read())

# Fedimg Consumer configurations
PROCESS_COUNT = config['general']['process_count']
STATUS_FILTER = ('FINISHED_INCOMPLETE', 'FINISHED')

ACTIVE_SERVICES = config['general']['active_services']
CLEAN_UP_ON_FAILURE = config['general']['clean_up_on_failure']
DELETE_IMAGES_ON_FAILURE = config['general']['delete_images_on_failure']

AWS_UTIL_USER = config.get('aws', {}).get('util_username', {})
AWS_TEST_USER = config.get('aws', {}).get('test_username', {})
AWS_ACCESS_ID = config.get('aws', {}).get('access_id')
AWS_SECRET_KEY = config.get('aws', {}).get('secret_key')
AWS_VOLUME_SIZE = config.get('aws', {}).get('volume_size')
AWS_VOLUME_TYPES = config.get('aws', {}).get('volume_types')
AWS_VOLUME_VIA_S3 = config.get('aws', {}).get('volume_via_s3')
AWS_IAM_PROFILE = config.get('aws', {}).get('iam_profile')
AWS_REGIONS = config.get('aws', {}).get('regions', {})
AWS_ROOT_VOLUME_SIZE = config.get('aws', {}).get('root_volume_size', {})
AWS_BASE_REGION = config.get('aws', {}).get('base_region', {})
AWS_DELETE_RESOURCES = config.get('aws', {}).get('delete_resources', True)



# This file is part of fedimg.
# Copyright (C) 2014-2015 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#           Ralph Bean <rbean@redhat.com>
#

import logging
log = logging.getLogger("fedmsg")

from time import sleep

import paramiko
from libcloud.compute.base import NodeImage
from libcloud.compute.deployment import MultiStepDeployment
from libcloud.compute.deployment import ScriptDeployment, SSHKeyDeployment
from libcloud.compute.providers import get_driver
from libcloud.compute.types import DeploymentException
from libcloud.compute.types import KeyPairDoesNotExistError

import fedimg
import fedimg.messenger
from fedimg.util import get_file_arch
from fedimg.util import region_to_driver, ssh_connection_works


class EC2ServiceException(Exception):
    """ Custom exception for EC2Service. """
    pass


class EC2UtilityException(EC2ServiceException):
    """ Something went wrong with writing the image file to a volume with the
        utility instance. """
    pass


class EC2AMITestException(EC2ServiceException):
    """ Something went wrong when a newly-registered AMI was tested. """
    pass


class EC2Service(object):
    """ An object for interacting with an EC2 upload process.
        Takes a URL to a raw.xz image. """

    def __init__(self, raw_url, virt_type='hvm', vol_type='standard'):

        self.raw_url = raw_url
        self.virt_type = virt_type
        self.vol_type = vol_type
        # All of these are set to appropriate values throughout
        # the upload process.
        self.util_node = None
        self.util_volume = None
        self.images = []
        self.snapshot = None
        self.test_node = None

        self.destination = ''

        # It's possible that these values will never change
        self.test_success = False
        self.dup_count = 0  # counter: helps avoid duplicate AMI names

        # Will be lists of dicts containing AMI info
        self.util_amis = []
        self.test_amis = []

        # Populate list of AMIs by reading the AMI details from the config file
        for line in fedimg.AWS_AMIS.split('\n'):
            """ AWS_AMIS lines have pipe-delimited attrs at these indicies:
            0: region (ex. eu-west-1)
            1: OS (ex. RHEL)
            2: version (ex. 5.7)
            3: arch (ex. x86_64)
            4: ami name (ex. ami-68e3d32d) """

            # strip line to avoid any newlines or spaces from sneaking in
            attrs = line.strip().split('|')

            # old configuration
            if len(attrs)==6:

                info = {'region': attrs[0],
                        'driver': region_to_driver(attrs[0]),
                        'os': attrs[1],
                        'ver': attrs[2],
                        'arch': attrs[3],
                        'ami': attrs[4],
                        'aki': attrs[5]}

            # new configuration
            elif len(attrs)==4:

                info = {'region': attrs[0],
                        'driver': region_to_driver(attrs[0]),
                        'arch': attrs[1],
                        'ami': attrs[2],
                        'aki': attrs[3]}



            # For now, read in all AMIs to these lists, and narrow
            # down later. TODO: This could be made a bit nicer...
            self.util_amis.append(info)
            self.test_amis.append(info)

        # Get file name, build name, a description, and the image arch
        # all from the .raw.xz file name.
        self.file_name = self.raw_url.split('/')[-1]
        self.build_name = self.file_name.replace('.raw.xz', '')
        self.image_desc = "Created from build {0}".format(self.build_name)
        self.image_arch = get_file_arch(self.file_name)

        # Filter the AMI lists appropriately
        # (no EBS-enabled instance types offer a 32 bit architecture, and we
        # need EBS for registration on the utility instance, so they must be
        # x86_64)
        self.util_amis = [a for a in self.util_amis
                          if a['arch'] == 'x86_64']
        self.test_amis = [a for a in self.test_amis
                          if a['arch'] == self.image_arch]

    def _clean_up(self, driver, delete_images=False):
        """ Cleans up resources via a libcloud driver. """
        log.info('Cleaning up resources')
        if delete_images and len(self.images) > 0:
            for image in self.images:
                driver.delete_image(image)

        if self.snapshot and len(self.images) == 0:
            driver.destroy_volume_snapshot(self.snapshot)
            self.snapshot = None

        if self.util_node:
            driver.destroy_node(self.util_node)
            # Wait for node to be terminated
            while ssh_connection_works(fedimg.AWS_UTIL_USER,
                                       self.util_node.public_ips[0],
                                       fedimg.AWS_KEYPATH):
                sleep(10)
            self.util_node = None
        if self.util_volume:
            # Destroy /dev/sdb or whatever
            driver.destroy_volume(self.util_volume)
            self.util_volume = None
        if self.test_node:
            driver.destroy_node(self.test_node)
            self.test_node = None

    def upload(self, compose_meta):
        """ Registers the image in each EC2 region. """

        log.info('EC2 upload process started')

        # Get a starting utility AMI in some region to use as an origin
        ami = self.util_amis[0]  # Select the starting AMI to begin
        self.destination = 'EC2 ({region})'.format(region=ami['region'])

        fedimg.messenger.message('image.upload', self.raw_url,
                                 self.destination, 'started',
                                 compose=compose_meta)

        try:
            # Connect to the region through the appropriate libcloud driver
            cls = ami['driver']
            driver = cls(fedimg.AWS_ACCESS_ID, fedimg.AWS_SECRET_KEY)

            # select the desired node attributes
            sizes = driver.list_sizes()
            reg_size_id = 'm1.xlarge'

            # check to make sure we have access to that size node
            # TODO: Add try/except if for some reason the size isn't
            # available?
            size = [s for s in sizes if s.id == reg_size_id][0]
            base_image = NodeImage(id=ami['ami'], name=None, driver=driver)

            # Name the utility node
            name = 'Fedimg AMI builder'

            # Block device mapping for the utility node
            # (Requires this second volume to write the image to for
            # future registration.)
            mappings = [{'VirtualName': None,  # cannot specify with Ebs
                         'Ebs': {'VolumeSize': fedimg.AWS_UTIL_VOL_SIZE,
                                 'VolumeType': self.vol_type,
                                 'DeleteOnTermination': 'false'},
                         'DeviceName': '/dev/sdb'}]

            # Read in the SSH key
            with open(fedimg.AWS_PUBKEYPATH, 'rb') as f:
                key_content = f.read()

            # Add key to authorized keys for root user
            step_1 = SSHKeyDeployment(key_content)

            # Add script for deployment
            # Device becomes /dev/xvdb on instance
            script = "touch test"  # this isn't so important for the util inst.
            step_2 = ScriptDeployment(script)

            # Create deployment object (will set up SSH key and run script)
            msd = MultiStepDeployment([step_1, step_2])

            log.info('Deploying utility instance')
            while True:
                try:
                    self.util_node = driver.deploy_node(
                        name=name,
                        image=base_image,
                        size=size,
                        ssh_username=fedimg.AWS_UTIL_USER,
                        ssh_alternate_usernames=[''],
                        ssh_key=fedimg.AWS_KEYPATH,
                        deploy=msd,
                        kernel_id=ami['aki'],
                        ex_metadata={'build':
                                     self.build_name},
                        ex_keyname=fedimg.AWS_KEYNAME,
                        ex_security_groups=['ssh'],
                        ex_ebs_optimized=True,
                        ex_blockdevicemappings=mappings)

                except KeyPairDoesNotExistError:
                    # The keypair is missing from the current region.
                    # Let's install it and try again.
                    log.exception('Adding missing keypair to region')
                    driver.ex_import_keypair(fedimg.AWS_KEYNAME,
                                             fedimg.AWS_PUBKEYPATH)
                    continue

                except Exception as e:
                    # We might have an invalid security group, aka the 'ssh'
                    # security group doesn't exist in the current region. The
                    # reason this is caught here is because the related
                    # exception that prints`InvalidGroup.NotFound is, for
                    # some reason, a base exception.
                    if 'InvalidGroup.NotFound' in str(e):
                        log.exception('Adding missing security'
                                      'group to region')
                        # Create the ssh security group
                        driver.ex_create_security_group('ssh', 'ssh only')
                        driver.ex_authorize_security_group('ssh', '22', '22',
                                                           '0.0.0.0/0')
                        continue
                    else:
                        raise
                break

            # Wait until the utility node has SSH running
            while not ssh_connection_works(fedimg.AWS_UTIL_USER,
                                           self.util_node.public_ips[0],
                                           fedimg.AWS_KEYPATH):
                sleep(10)

            log.info('Utility node started with SSH running')

            # Connect to the utility node via SSH
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(self.util_node.public_ips[0],
                           username=fedimg.AWS_UTIL_USER,
                           key_filename=fedimg.AWS_KEYPATH)

            # Curl the .raw.xz file down from the web, decompressing it
            # and writing it to the secondary volume defined earlier by
            # the block device mapping.
            # curl with -L option, so we follow redirects
            cmd = "sudo sh -c 'curl -L {0} | xzcat > /dev/xvdb'".format(
                  self.raw_url)
            chan = client.get_transport().open_session()
            chan.get_pty()  # Request a pseudo-term to get around require tty

            log.info('Executing utility script')

            # Run the above command and wait for its exit status
            chan.exec_command(cmd)
            status = chan.recv_exit_status()
            if status != 0:
                # There was a problem with the SSH command
                log.error('Problem writing volume with utility instance')

                data = "(no data)"
                if chan.recv_ready():
                    data = chan.recv(1024 * 32)

                fedimg.messenger.message('image.upload', self.raw_url,
                                         self.destination, 'failed',
                                         extra={'data': data},
                                         compose=compose_meta)

                raise EC2UtilityException(
                    "Problem writing image to utility instance volume. "
                    "Command exited with status {0}.\n"
                    "command: {1}\n"
                    "output: {2}".format(status, cmd, data))

            client.close()

            # Get volume name that image was written to
            vol_id = [x['ebs']['volume_id'] for x in
                      self.util_node.extra['block_device_mapping'] if
                      x['device_name'] == '/dev/sdb'][0]

            log.info('Destroying utility node')

            # Terminate the utility instance
            driver.destroy_node(self.util_node)

            # Wait for utility node to be terminated
            while ssh_connection_works(fedimg.AWS_UTIL_USER,
                                       self.util_node.public_ips[0],
                                       fedimg.AWS_KEYPATH):
                sleep(10)

            # Wait a little longer since loss of SSH connectivity doesn't mean
            # that the node's destroyed
            # TODO: Check instance state rather than this lame sleep thing
            sleep(45)

            # Take a snapshot of the volume the image was written to
            self.util_volume = [v for v in driver.list_volumes()
                                if v.id == vol_id][0]
            snap_name = 'fedimg-snap-{0}'.format(self.build_name)

            log.info('Taking a snapshot of the written volume')

            self.snapshot = driver.create_volume_snapshot(self.util_volume,
                                                          name=snap_name)
            snap_id = str(self.snapshot.id)

            while self.snapshot.extra['state'] != 'completed':
                # Re-obtain snapshot object to get updates on its state
                self.snapshot = [s for s in driver.list_snapshots()
                                 if s.id == snap_id][0]
                sleep(10)

            # Make the snapshot public, so that the AMIs can be copied
            is_snapshot_public = False
            while True:
                is_snapshot_public = driver.ex_modify_snapshot_attribute(self.snapshot, {
                    'CreateVolumePermission.Add.1.Group': 'all'
                })
                if is_snapshot_public:
                    break

                log.info('Snapshot is not public yet. Retry in 20')
                sleep(20)

            log.info('Snapshot taken & made public')

            # Delete the volume now that we've got the snapshot
            driver.destroy_volume(self.util_volume)
            # make sure Fedimg knows that the vol is gone
            self.util_volume = None

            log.info('Destroyed volume')

            # Actually register image
            log.info('Registering image as an AMI')

            if self.virt_type == 'paravirtual':
                image_name = "{0}-{1}-PV-{2}-0".format(self.build_name,
                                                  ami['region'],
                                                  self.vol_type)
                test_size_id = 'm1.xlarge'
                # test_amis will include AKIs of the appropriate arch
                registration_aki = [a['aki'] for a in self.test_amis
                                    if a['region'] == ami['region']][0]
                reg_root_device_name = '/dev/sda'
            else:  # HVM
                image_name = "{0}-{1}-HVM-{2}-0".format(self.build_name,
                                                    ami['region'],
                                                    self.vol_type)
                test_size_id = 'm3.2xlarge'
                # Can't supply a kernel image with HVM
                registration_aki = None
                reg_root_device_name = '/dev/sda1'

            # For this block device mapping, we have our volume be
            # based on the snapshot's ID
            mapping = [{'DeviceName': reg_root_device_name,
                        'Ebs': {'SnapshotId': snap_id,
                                'VolumeSize': fedimg.AWS_TEST_VOL_SIZE,
                                'VolumeType': self.vol_type,
                                'DeleteOnTermination': 'true'}}]

            # Avoid duplicate image name by incrementing the number at the
            # end of the image name if there is already an AMI with that name.
            # TODO: This process could be written nicer.
            while True:
                try:
                    if self.dup_count > 0:
                        # Remove trailing '-0' or '-1' or '-2' or...
                        image_name = '-'.join(image_name.split('-')[:-1])
                        # Re-add trailing dup number with new count
                        image_name += '-{0}'.format(self.dup_count)
                    # Try to register with that name
                    self.images.append(driver.ex_register_image(
                        image_name,
                        description=self.image_desc,
                        root_device_name=reg_root_device_name,
                        block_device_mapping=mapping,
                        virtualization_type=self.virt_type,
                        kernel_id=registration_aki,
                        architecture=self.image_arch))
                except Exception as e:
                    # Check if the problem was a duplicate name
                    if 'InvalidAMIName.Duplicate' in str(e):
                        # Keep trying until an unused name is found
                        self.dup_count += 1
                        continue
                    else:
                        raise
                break

            log.info('Completed image registration')

            # Emit success fedmsg
            # TODO: Can probably move this into the above try/except,
            # to avoid just dumping all the messages at once.
            for image in self.images:
                fedimg.messenger.message('image.upload', self.raw_url,
                                         self.destination, 'completed',
                                         extra={'id': image.id,
                                                'virt_type': self.virt_type,
                                                'vol_type': self.vol_type},
                                         compose=compose_meta)

            # Now, we'll spin up a node of the AMI to test:

            # Add script for deployment
            # Device becomes /dev/xvdb on instance
            script = "touch test"
            step_2 = ScriptDeployment(script)

            # Create deployment object
            msd = MultiStepDeployment([step_1, step_2])

            log.info('Deploying test node')

            # Pick a name for the test instance
            name = 'Fedimg AMI tester'

            # Select the appropriate size for the instance
            size = [s for s in sizes if s.id == test_size_id][0]

            # Alert the fedmsg bus that an image test is starting
            fedimg.messenger.message('image.test', self.raw_url,
                                     self.destination, 'started',
                                     extra={'id': self.images[0].id,
                                            'virt_type': self.virt_type,
                                            'vol_type': self.vol_type},
                                     compose=compose_meta)

            # Actually deploy the test instance
            try:
                self.test_node = driver.deploy_node(
                    name=name, image=self.images[0], size=size,
                    ssh_username=fedimg.AWS_TEST_USER,
                    ssh_alternate_usernames=['root'],
                    ssh_key=fedimg.AWS_KEYPATH,
                    deploy=msd,
                    kernel_id=registration_aki,
                    ex_metadata={'build': self.build_name},
                    ex_keyname=fedimg.AWS_KEYNAME,
                    ex_security_groups=['ssh'],
                    )
            except Exception as e:
                fedimg.messenger.message('image.test', self.raw_url,
                                         self.destination, 'failed',
                                         extra={'id': self.images[0].id,
                                                'virt_type': self.virt_type,
                                                'vol_type': self.vol_type},
                                         compose=compose_meta)

                raise EC2AMITestException("Failed to boot test node %r." % e)

            # Wait until the test node has SSH running
            while not ssh_connection_works(fedimg.AWS_TEST_USER,
                                           self.test_node.public_ips[0],
                                           fedimg.AWS_KEYPATH):
                sleep(10)

            log.info('Starting AMI tests')

            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            client.connect(self.test_node.public_ips[0],
                           username=fedimg.AWS_TEST_USER,
                           key_filename=fedimg.AWS_KEYPATH)

            # Run /bin/true on the test instance as a simple "does it
            # work" test
            cmd = "/bin/true"
            chan = client.get_transport().open_session()
            chan.get_pty()  # Request a pseudo-term to get around requiretty

            log.info('Running AMI test script')

            chan.exec_command(cmd)

            # Again, wait for the test command's exit status
            if chan.recv_exit_status() != 0:
                # There was a problem with the SSH command
                log.error('Problem testing new AMI')

                data = "(no data)"
                if chan.recv_ready():
                    data = chan.recv(1024 * 32)

                fedimg.messenger.message('image.test', self.raw_url,
                                         self.destination, 'failed',
                                         extra={'id': self.images[0].id,
                                                'virt_type': self.virt_type,
                                                'vol_type': self.vol_type,
                                                'data': data},
                                         compose=compose_meta)

                raise EC2AMITestException("Tests on AMI failed.\n"
                                          "output: %s" % data)

            client.close()

            log.info('AMI test completed')
            fedimg.messenger.message('image.test', self.raw_url,
                                     self.destination, 'completed',
                                     extra={'id': self.images[0].id,
                                            'virt_type': self.virt_type,
                                            'vol_type': self.vol_type},
                                     compose=compose_meta)

            # Let this EC2Service know that the AMI test passed, so
            # it knows how to proceed.
            self.test_success = True

            log.info('Destroying test node')

            # Destroy the test node
            driver.destroy_node(self.test_node)

            # Make AMIs public
            for image in self.images:
                driver.ex_modify_image_attribute(
                    image,
                    {'LaunchPermission.Add.1.Group': 'all'})

        except EC2UtilityException as e:
            log.exception("Failure")
            if fedimg.CLEAN_UP_ON_FAILURE:
                self._clean_up(driver,
                               delete_images=fedimg.DELETE_IMAGES_ON_FAILURE)
            return 1

        except EC2AMITestException as e:
            log.exception("Failure")
            if fedimg.CLEAN_UP_ON_FAILURE:
                self._clean_up(driver,
                               delete_images=fedimg.DELETE_IMAGES_ON_FAILURE)
            return 1

        except DeploymentException as e:
            log.exception("Problem deploying node: {0}".format(e.value))
            if fedimg.CLEAN_UP_ON_FAILURE:
                self._clean_up(driver,
                               delete_images=fedimg.DELETE_IMAGES_ON_FAILURE)
            return 1

        except Exception as e:
            # Just give a general failure message.
            log.exception("Unexpected exception")
            if fedimg.CLEAN_UP_ON_FAILURE:
                self._clean_up(driver,
                               delete_images=fedimg.DELETE_IMAGES_ON_FAILURE)
            return 1

        else:
            self._clean_up(driver)

        if self.test_success:
            # Copy the AMI to every other region if tests passed
            copied_images = list()  # completed image copies (ami: image)

            # Use the AMI list as a way to cycle through the regions
            for ami in self.test_amis[1:]:  # we don't need the origin region

                # Choose an appropriate destination name for the copy
                alt_dest = 'EC2 ({region})'.format(
                    region=ami['region'])

                fedimg.messenger.message('image.upload',
                                         self.raw_url,
                                         alt_dest, 'started',
                                         compose=compose_meta)

                # Connect to the libcloud EC2 driver for the region we
                # want to copy into
                alt_cls = ami['driver']
                alt_driver = alt_cls(fedimg.AWS_ACCESS_ID,
                                     fedimg.AWS_SECRET_KEY)

                # Construct the full name for the image copy
                if self.virt_type == 'paravirtual':
                    image_name = "{0}-{1}-PV-{2}-0".format(
                        self.build_name, ami['region'], self.vol_type)
                else:  # HVM
                    image_name = "{0}-{1}-HVM-{2}-0".format(
                        self.build_name, ami['region'], self.vol_type)

                log.info('AMI copy to {0} started'.format(ami['region']))

                # Avoid duplicate image name by incrementing the number at the
                # end of the image name if there is already an AMI with
                # that name.
                # TODO: Again, this could be written better
                while True:
                    try:
                        if self.dup_count > 0:
                            # Remove trailing '-0' or '-1' or '-2' or...
                            image_name = '-'.join(image_name.split('-')[:-1])
                            # Re-add trailing dup number with new count
                            image_name += '-{0}'.format(self.dup_count)

                        # Actually run the image copy from the origin region
                        # to the current region.
                        for image in self.images:
                            image_copy = alt_driver.copy_image(
                                image,
                                self.test_amis[0]['region'],
                                name=image_name,
                                description=self.image_desc)
                            # Add the image copy to a list so we can work with
                            # it later.
                            copied_images.append(image_copy)

                            log.info('AMI {0} copied to AMI {1}'.format(
                                image, image_name))

                    except Exception as e:
                        # Check if the problem was a duplicate name
                        if 'InvalidAMIName.Duplicate' in str(e):
                            # Keep trying until an unused name is found.
                            # This probably won't trigger, since it seems
                            # like EC2 doesn't mind duplicate AMI names
                            # when they are being copied, only registered.
                            # Strange, but apprently true.
                            self.dup_count += 1
                            continue
                        else:
                            # TODO: Catch a more specific exception
                            log.exception(
                                'Image copy to {0} failed'.format(
                                    ami['region']))
                            fedimg.messenger.message('image.upload',
                                                     self.raw_url,
                                                     alt_dest, 'failed',
                                                     compose=compose_meta)
                    break

            # Now cycle through and make all of the copied AMIs public
            # once the copy process has completed. Again, use the test
            # AMI list as a way to have region and arch data:

            # We don't need the origin region, since the AMI was made there:
            self.test_amis = self.test_amis[1:]

            for image in copied_images:
                ami = self.test_amis[copied_images.index(image)]
                alt_cls = ami['driver']
                alt_driver = alt_cls(fedimg.AWS_ACCESS_ID,
                                     fedimg.AWS_SECRET_KEY)

                # Get an appropriate name for the region in question
                alt_dest = 'EC2 ({region})'.format(region=ami['region'])

                # Need to wait until the copy finishes in order to make
                # the AMI public.
                is_image_public = False
                while True:
                    try:
                        # Make the image public
                        is_image_public = alt_driver.ex_modify_image_attribute(
                            image,
                            {'LaunchPermission.Add.1.Group': 'all'}
                        )
                    except Exception as e:
                        if 'InvalidAMIID.Unavailable' in str(e):
                            # The copy isn't done, so wait 20 seconds
                            # and try again.
                            sleep(20)
                            continue
                    break

                if is_image_public:
                    log.info('Made {0} public ({1}, {2}, {3}, {4})'.format(
                        image.id, self.build_name, self.virt_type,
                        self.vol_type, ami['region']))
                else:
                    log.info('{0} is private ({1}, {2}, {3}, {4})'.format(
                        image.id, self.build_name, self.virt_type,
                        self.vol_type, ami['region']))

                # Make the snapshot for the image public.
                is_snapshot_public = False
                snapshot = None

                alt_ami = alt_driver.get_image(image.id)
                blk_device_mapping = alt_ami.extra['block_device_mapping']
                if len(blk_device_mapping) == 1:
                    snapshot_id = blk_device_mapping[0]['ebs']['snapshot_id']

                    # The `list_snapshots` method requires a snapshot object.
                    # which then fetches the id of the snapshot and fetches the
                    # detail of the snapshot. So, I am making an empty snapshot
                    # object here and attaching the value to the `id` attribute
                    # so that the list_snapshots method just works
                    snapshot_obj = type('', (), {})()
                    snapshot_obj.id = snapshot_id
                    snapshot = alt_driver.list_snapshots(snapshot=snapshot_obj)

                if snapshot is not None:
                    snapshot = snapshot[0]
                    while True:
                        is_snapshot_public = (
                            alt_driver.ex_modify_snapshot_attribute(
                                snapshot, {
                                    'CreateVolumePermission.Add.1.Group': 'all'
                                })
                            )
                        if is_snapshot_public:
                            break

                        log.info('Snapshot is not public yet. Retry in 20')
                        sleep(20)
                else:
                    is_snapshot_public = False
                    log.info('Search (%s, %s) returned no results' % (
                        snapshot_id, ami['region']))

                if is_snapshot_public:
                    log.info('Snapshot (%s, %s) made public' % (
                        snapshot_id, ami['region']))
                else:
                    log.info('Snapshot (%s, %s) still private' % (
                        snapshot_id, ami['region']))

                fedimg.messenger.message('image.upload',
                                         self.raw_url,
                                         alt_dest, 'completed',
                                         extra={'id': image.id,
                                                'virt_type': self.virt_type,
                                                'vol_type': self.vol_type},
                                         compose=compose_meta)

            return 0



# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
#

import logging
log = logging.getLogger("fedmsg")

import re

from time import sleep

import fedimg.messenger

from fedimg.utils import external_run_command, get_item_from_regex
from fedimg.utils import get_volume_type_from_image
from fedimg.utils import get_virt_type_from_image
from fedimg.utils import get_image_name_from_ami_name
from fedimg.services.ec2.ec2base import EC2Base


class EC2ImagePublisher(EC2Base):
    """ Comment goes here """

    def __init__(self, **kwargs):
        defaults = {
            'access_key': None,
            'compose_id': None,
            'image_id': None,
            'image_name': 'Fedora-AMI',
            'image_description': 'Fedora AMI Description',
            'service': 'EC2',
            'region': None,
            'secret_key': None,
            'visibility': 'all',
            'push_notifications': False,
        }

        for (prop, default) in defaults.iteritems():
            setattr(self, prop, kwargs.get(prop, default))

    def _retry_till_image_is_public(self, image):
        """ Comment goes here """

        driver = self._connect()

        is_image_public = False
        while True:
            try:
                is_image_public = driver.ex_modify_image_attribute(
                    image,
                    {'LaunchPermission.Add.1.Group': 'all'})
            except Exception as e:
                if 'InvalidAMIID.Unavailable' in str(e):
                    # The copy isn't completed yet, so wait for 20 seconds
                    # more.
                    sleep(20)
                    continue
            break

        return is_image_public

    def _retry_till_snapshot_is_public(self, snapshot):

        driver = self._connect()

        while True:
            is_snapshot_public = driver.ex_modify_snapshot_attribute(
                snapshot,
                {'CreateVolumePermission.Add.1.Group': 'all'})

            if is_snapshot_public:
                break

        return is_snapshot_public


    def _retry_till_snapshot_is_available(self, image):

        driver = self._connect()
        while True:
            image = driver.get_image(image.id)
            snapshot_id = image.extra['block_device_mapping'][0]['ebs']['snapshot_id']

            if snapshot_id:
                break

        return snapshot_id

    def _generate_dummy_snapshot_object(self, snapshot_id):

        driver = self._connect()

        snapshot_obj = type('', (), {})()
        snapshot_obj.id = snapshot_id
        snapshot = driver.list_snapshots(snapshot=snapshot_obj)

        return snapshot

    def _retry_till_blk_mapping_is_available(self, image):

        while True:
            image = self._connect().get_image(image_id=image.id)
            blk_mapping = image.extra['block_device_mapping']

            if blk_mapping:
                return blk_mapping

    def get_snapshot_from_image(self, image):
        """ Comment goes here """
        if isinstance(image, str):
            image_id = image
            image = self._connect().get_image(image_id)

        blk_mapping = image.extra['block_device_mapping']
        if not blk_mapping:
            blk_mapping = self._retry_till_blk_mapping_is_available(image)

        snapshot_id = blk_mapping[0]['ebs']['snapshot_id']
        if snapshot_id is None:
            snapshot_id = self._retry_till_snapshot_is_available(image)

        snapshot = self._generate_dummy_snapshot_object(snapshot_id)[0]

        return snapshot

    def get_volume_type_from_image(self, image):
        if isinstance(image, str):
            image_id = image
            image = self._connect().get_image(image_id)

        blk_mapping = image.extra['block_device_mapping']
        if not blk_mapping:
            blk_mapping = self._retry_till_blk_mapping_is_available(image)

        return blk_mapping[0]['ebs']['volume_type']

    def get_virt_type_from_image(self, image):
        if isinstance(image, str):
            image_id = image
            image = self._connect().get_image(image_id)

        blk_mapping = image.extra['block_device_mapping']
        if not blk_mapping:
            blk_mapping = self._retry_till_blk_mapping_is_available(image)
 
        device_name = blk_mapping[0]['ebs']['volume_type']

        if device_name.endswith('sda1'):
            return 'hvm'
        else:
            return 'paravirtual'

    def publish_images(self, region_image_mapping=None):
        """ Comment goes here """

        published_images = []
        if region_image_mapping is None:
            return published_images

        for region, image_id in region_image_mapping:
            self.set_region(region)

            log.info('Publish image (%s) in %s started' % (image_id, region))
            image = self._connect().get_image(image_id=image_id)
            is_image_public = self._retry_till_image_is_public(image)
            log.info('Publish image (%s) in %s completed' % (image_id, region))

            log.info('Publish snaphsot for image (%s) in %s started' % (image_id, region))
            snapshot = self.get_snapshot_from_image(image)
            log.info('Fetched snapshot for image (%s): %s' % (image_id, snapshot.id))
            is_snapshot_public = self._retry_till_snapshot_is_public(snapshot)
            log.info('Publish snaphsot for image (%s) in %s completed' % (image_id, region))

            volume_type = self.get_volume_type_from_image(image)
            virt_type = self.get_virt_type_from_image(image)

            if self.push_notifications:
                fedimg.messenger.notify(
                    topic='image.publish',
                    msg=dict(
                        image_name=image.name,
                        destination=self.region,
                        service=self.service,
                        compose=self.compose_id,
                        extra=dict(
                            id=image.id,
                            virt_type=virt_type,
                            vol_type=volume_type
                        )
                    )
                )

            published_images.append({
                'image_id': image.id,
                'is_image_public': is_image_public,
                'snapshot_id': snapshot.id,
                'is_snapshot_public': is_snapshot_public,
                'regions': self.region
            })

        return published_images

    def copy_images_to_regions(self, image_id=None, base_region=None, regions=None):
        """ Comment goes here """

        if (image_id is None) or (regions is None) or (base_region is None):
            return

        counter = 0
        copied_images = []

        self.set_region(base_region)
        image = self._connect().get_image(image_id=image_id)
        if not image:
            return []

        for region in regions:
            log.info('Copy %s to %s started' % (image_id, region))
            self.set_region(region)
            self.image_name = get_image_name_from_ami_name(image.name, region)

            while True:
                if counter > 0:
                    self.image_name = re.sub(
                        '\d(?!\d)',
                        lambda x: str(int(x.group(0))+1),
                        self.image_name
                    )
                try:
                    copied_image = self._connect().copy_image(
                        source_region=base_region,
                        image=image,
                        name=self.image_name,
                        description=self.image_description)

                    virt_type = image.extra['virtualization_type']
                    volume_type = image.extra['block_device_mapping'][0]['ebs']['volume_type']

                    if self.push_notifications:
                        fedimg.messenger.notify(
                            topic='image.copy',
                            msg=dict(
                                image_name=copied_image.name,
                                destination=self.region,
                                service=self.service,
                                compose_id=self.compose_id,
                                extra=dict(
                                    virt_type=virt_type,
                                    vol_type=volume_type
                                )
                            )
                        )

                    log.info('Copy %s to %s is completed.' % (image_id, region))
                    copied_images.append({
                        'region': region,
                        'copied_image_id': copied_image.id
                    })
                    break

                except Exception as e:
                    log.info('Could not register '
                             'with name: %r' % self.image_name)
                    if 'InvalidAMIName.Duplicate' in str(e):
                        counter = counter + 1
                    else:
                        log.info('Failed')
                        break

        return copied_images

    def deprecate_images(self, image_ids=None, snapshot_perm='all'):
        raise NotImplementedError

    def delete_images(self, image_ids=None, snapshot_perm='all'):
        raise NotImplementedError



# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#

import socket
hostname = socket.gethostname()

NUM_BASE_THREADS = 4
NUM_ATOMIC_THREADS = 2
NUM_PORTS = 2 * ((NUM_BASE_THREADS + NUM_ATOMIC_THREADS) + 1)

config = {
    'fedimgconsumer.dev.enabled': True,
    'fedimgconsumer.prod.enabled': False,
    'fedimgconsumer.stg.enabled': False,
    'endpoints': {
        "fedimg.%s" % hostname: [
            "tcp://127.0.0.1:60%0.2i" % (i)
            for i in range(NUM_PORTS)
        ],
    },
}



# -*- coding: utf-8 -*-
# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#           Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
"""
This is the `fedmsg consumer`_ that subscribes to the topic emitted after the
completion of the nightly and production compose. The consumer on receving the
message uploads the image using the API of the cloud providers.
"""

import logging
import multiprocessing.pool

import fedmsg.consumers
import fedmsg.encoding
import fedfind.release

import fedimg.uploader

from fedimg.config import PROCESS_COUNT, STATUS_FILTER
from fedimg.utils import get_rawxz_urls, get_value_from_dict

LOG = logging.getLogger(__name__)


class FedimgConsumer(fedmsg.consumers.FedmsgConsumer):
    """
    A `fedmsg consumer`_ that listens to the pungi compose topics and kicks
    of the process to upload the images to various cloud providers.

    Attributes:
        topic (str): The topics this consumer is subscribed to. Set to
            ``org.fedoraproject.prod.pungi.compose.status.change``.
        config_key (str): The key to set to ``True`` in the fedmsg config to
            enable this consumer. The key is ``fedimgconsumer.prod.enabled``.
    """
    topic = ['org.fedoraproject.prod.pungi.compose.status.change']
    config_key = "fedimgconsumer.prod.enabled"

    def __init__(self, *args, **kwargs):
        LOG.info("FedimgConsumer initializing")
        super(FedimgConsumer, self).__init__(*args, **kwargs)

        # Threadpool for upload jobs
        LOG.info("Creating thread pool of %s process", PROCESS_COUNT)
        self.upload_pool = multiprocessing.pool.ThreadPool(
            processes=PROCESS_COUNT
        )
        LOG.info("FedimgConsumer initialized")

    def consume(self, msg):
        """
        This is called when we receive a message matching our topics.

        Args:
            msg (dict): The raw message from fedmsg.
        """
        LOG.info('Received %r %r', msg['topic'], msg['body']['msg_id'])

        msg_info = msg['body']['msg']
        if msg_info['status'] not in STATUS_FILTER:
            return

        location = msg_info['location']
        compose_id = msg_info['compose_id']
        compose_metadata = fedfind.release.get_release_cid(compose_id).metadata
        images_meta = get_value_from_dict(
            compose_metadata,
            'images',
            'payload',
            'images',
            'CloudImages',
            'x86_64'
        )

        if images_meta is None:
            LOG.debug('No compatible image found to process')
            return

        upload_urls = get_rawxz_urls(location, images_meta)
        if len(upload_urls) > 0:
            LOG.info("Start processing compose id: %s", compose_id)
            fedimg.uploader.upload(
                pool=self.upload_pool,
                urls=upload_urls,
                compose_id=compose_id
            )


class FedimgStagingConsumer(FedimgConsumer):
    """
    A `fedmsg consumer`_ that listens to the staging pungi compose topics and
    kicks of the process to upload the images to various cloud providers.

    Attributes:
        topic (str): The topics this consumer is subscribed to. Set to
            ``org.fedoraproject.stg.pungi.compose.status.change``.
        config_key (str): The key to set to ``True`` in the fedmsg config to
            enable this consumer. The key is ``fedimgconsumer.stg.enabled``.
    """
    topic = ['org.fedoraproject.stg.pungi.compose.status.change']
    config_key = "fedimgconsumer.stg.enabled"


class FedimgDevConsumer(FedimgConsumer):
    """
    A `fedmsg consumer`_ that listens to the dev pungi compose topics and
    kicks of the process to upload the images to various cloud providers.

    Attributes:
        topic (str): The topics this consumer is subscribed to. Set to
            ``org.fedoraproject.dev.pungi.compose.status.change``.
        config_key (str): The key to set to ``True`` in the fedmsg config to
            enable this consumer. The key is ``fedimgconsumer.dev.enabled``.
    """
    topic = ['org.fedoraproject.dev.pungi.compose.status.change']
    config_key = "fedimgconsumer.dev.enabled"



# This file is part of fedimg.
# Copyright (C) 2014-2017 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  Sayan Chowdhury <sayanchowdhury@fedoraproject.org>
#

import logging
log = logging.getLogger("fedmsg")

import re

import fedimg.messenger

from fedimg.utils import external_run_command, get_item_from_regex
from fedimg.config import AWS_DELETE_RESOURCES
from fedimg.services.ec2.ec2base import EC2Base


class EC2ImageUploader(EC2Base):
    """
    The 'ec2.ec2imguploader.EC2ImageUploader` creates the Amazon Machine Image
    (AMI) from the source.

    Args:
        access_key (str): Access key for the account.
        availability_zone (str): Availability zone to use
        compose_id (str): ID of the Compose
        image_name (str): Name of the Image
        image_description (str): Description of the image
        image_virtualization_type (str): Virtulization Type of the Image
        image_architecture (str): Architecture of the Image
        image_url (str): URL of the the Image
        image_volume_type (str): Volume Type of the Image
        image_format (str): Format of the Image
        region (str): Region to upload
        service (str): Name of the Fedimg service.
        secret_key (str): Secret Key for the account
        s3_bucket_name (str): Name of the S3 bucket
        volume_via_s3 (bool): Bool to control AMI using S3-method
        root_volume_size (str): Size of the AWS volume
        push_notifications (bool): Push notifications to fedmsg
    """

    def __init__(self, *args, **kwargs):
        defaults = {
            'access_key': None,
            'availability_zone': None,
            'compose_id': None,
            'image_name': 'Fedora-AMI',
            'image_description': 'Fedora AMI Description',
            'image_virtualization_type': 'hvm',
            'image_architecture': 'x86_64',
            'image_url': None,
            'image_volume_type': 'gp2',
            'image_format': 'raw',
            'region': None,
            'service': 'EC2',
            'secret_key': None,
            's3_bucket_name': 'Fedora-S3-bucket',
            'volume_via_s3': True,
            'root_volume_size': 7,
            'push_notifications': False,
        }

        for (prop, default) in defaults.iteritems():
            setattr(self, prop, kwargs.get(prop, default))

    def _determine_root_device_name(self):
        root_device_name = '/dev/sda'
        if self.image_virtualization_type == 'hvm':
            root_device_name = '/dev/sda1'

        log.debug('Root device name is set to %r for %r' % (
            root_device_name, self.image_virtualization_type))

        return root_device_name

    def _create_block_device_map(self, snapshot):
        root_device_name = self._determine_root_device_name()
        block_device_map = {
            'DeviceName': root_device_name,
            'Ebs': {
                'SnapshotId': snapshot.id,
                'VolumeSize': self.root_volume_size,
                'VolumeType': self.image_volume_type,
                'DeleteOnTermination': True,
            }
        }

        log.debug('Block device map created for %s' % snapshot.id)

        return [block_device_map]

    def _retry_and_get_volume_id(self, task_id):
        while True:
            output, err, retcode = external_run_command([
                'euca-describe-conversion-tasks',
                task_id,
                '--region',
                self.region
            ])

            if 'completed' in output:
                log.debug('Task %r complete. Fetching volume id...' % task_id)
                match = re.search('\s(vol-\w{17})', output)
                volume_id = match.group(1)

                log.debug('The id of the created volume: %r' % volume_id)

                return volume_id

            log.debug('Failed to find complete. Task %r still running. '
                      'Sleeping for 10 seconds.' % task_id)

    def _create_snapshot(self, volume):
        snapshot = self._connect().create_volume_snapshot(
                volume=volume,
                name=self.image_name
        )
        snapshot_id = snapshot.id
        snapshot = self._retry_and_get_snapshot(snapshot_id)

        return snapshot

    def _retry_and_get_snapshot(self, snapshot_id):
        #FIXME: Rather that listing all snapshot. Add a patch to libcloud to
        # pull the snapshot using the snapshot id.
        snapshots = self._connect().list_snapshots()
        for snapshot in snapshots:
            if snapshot.id == snapshot_id:
                break

        while snapshot.extra['state'] != 'completed':
            snapshots = self._connect().list_snapshots()
            for snapshot in snapshots:
                if snapshot.id == snapshot_id:
                    break

        return snapshot

    def _create_volume(self, source):
        if self.volume_via_s3:
            output, err, retcode = external_run_command([
                'euca-import-volume',
                source,
                '-f',
                self.image_format,
                '--region',
                self.region,
                '-b',
                self.s3_bucket_name,
                '-z',
                self.availability_zone
            ])

            log.debug('Initiate task to upload the image via S3. '
                      'Fetching task id...')

            task_id = get_item_from_regex(output, regex='\s(import-vol-\w{8})')
            log.info('Fetched task_id: %r. Listening to the task.' % task_id)

            volume_id = self._retry_and_get_volume_id(task_id)

            volume = self.get_volume_from_volume_id(volume_id)
            log.info('Finish fetching volume object using volume_id')

            return volume

    def _register_image(self, snapshot):
        counter = 0
        block_device_map = self._create_block_device_map(snapshot)
        root_device_name = self._determine_root_device_name()
        while True:
            if counter > 0:
                self.image_name = re.sub('\d(?!\d)$',
                                         lambda x: str(int(x.group(0))+1),
                                         self.image_name)
            try:
                log.info('Registering the image in %r (snapshot id: %r) with '
                         'name %r' % (self.region, snapshot.id,
                                      self.image_name))
                image = self._connect().ex_register_image(
                    name=self.image_name,
                    description=self.image_description,
                    virtualization_type=self.image_virtualization_type,
                    architecture=self.image_architecture,
                    block_device_mapping=block_device_map,
                    root_device_name=root_device_name
                )

                if self.push_notifications:
                    fedimg.messenger.notify(
                        topic='image.upload',
                        msg=dict(
                            image_url=self.image_url,
                            image_name=self.image_name,
                            destination=self.region,
                            service=self.service,
                            status='completed',
                            compose=self.compose_id,
                            extra=dict(
                                id=image.id,
                                virt_type=self.image_virtualization_type,
                                vol_type=self.image_volume_type
                            )
                        )
                    )

                return image

            except Exception as e:
                log.info('Could not register with name: %r' % self.image_name)
                if 'InvalidAMIName.Duplicate' in str(e):
                    counter = counter + 1
                else:
                    raise

    def _remove_volume(self, volume):
        log.info('[CLEAN] Destroying volume: %r' % volume.id)
        self._connect().destroy_volume(volume)

    def clean_up(self, image_id, delete_snapshot=True, force=False):
        """
        Clean the resources created during the upload process

        Args:
            image_id (str): ID of the AMI
            delete_snapshot (bool): Bool to control snapshot deletion along
                                    with AMI.
            force: Override AWS_DELETE_RESOURCES config value

        Returns:
            Boolean: True, if resources are deleted else False
        """
        if not AWS_DELETE_RESOURCES and force:
            log.info('Deleting resource is disabled by config.'
                     'Override by passing force=True.')
            return False

        if self.volume:
            self._connect().destroy_volume(self.volume)

        self._connect().deregister_image(
            image_id,
            delete_snapshot=delete_snapshot
        )

        return True

    def set_image_virt_type(self, virt_type):
        """
        Set the `image_virtualization_type` attribute of the `EC2ImageUploader`
        object.

        Args:
            virt_type (str): virtualization type to set for the object.
        """
        self.image_virtualization_type = virt_type

    def set_image_url(self, image_url):
        """
        Set the `image_url` attribute of the `EC2ImageUploader` object.

        Args:
            image_url (str): image_url to set for the object.
        """
        self.image_url = image_url

    def set_image_name(self, image_name):
        """
        Set the `image_name` attribute of the `EC2ImageUploader` object.

        Args:
            image_name (str): image_name to set for the object.
        """
        self.image_name = image_name

    def set_image_volume_type(self, volume_type):
        """
        Set the `volume_type` attribute of the `EC2ImageUploader` object.

        Args:
            volume_type (str): volume_type to set for the object.
        """
        self.volume_type = volume_type

    def get_volume_from_volume_id(self, volume_id):
        """ Get the `` object from the volume_id.

        Args:
            volume_id (str): volume_id for the `EC2ImageUploader` object
        """
        #FIXME: This is not a optimized way of get all the volumes. Rather
        # send a patch to libcloud to filter the volume based on the volume_id

        volumes = self._connect().list_volumes()

        for volume in volumes:
            if volume.id == volume_id:
                return volume

    def set_availability_zone_for_region(self):
        """
        Returns a availability zone for the region
        """
        self.availability_zone = self._connect().ex_list_availability_zones(
            only_available=True)[0].name

    def create_volume(self, source):
        """
        Create a AWS volume out of the given source file

        Args:
            source (str): File path of the source file
        """
        log.info('Start creating the volume from source: %r' % source)
        return self._create_volume(source)

    def create_snapshot(self, source):
        """
        Creates a AWS snapshot out of the given source file

        Args:
            source (str): File path of the source file

        Returns:
            snapshot: `VolumeSnapshot` object
        """
        self.volume = self._create_volume(source)

        log.info('Start creating snapshot from volume: %r' % self.volume.id)
        snapshot = self._create_snapshot(self.volume)

        self._remove_volume(self.volume)

        return snapshot

    def register_image(self, snapshot):
        """
        Register the image for the given `` object,

        Args:
            snapshot: `VolumeSnapshot` object

        Returns:
            image: `NodeImage` object
        """
        image = self._register_image(snapshot)

        return image

    def create_image(self, source):
        """
        Create Amazon machin image out of the given source

        Args:
            source (str): path of the source cloud image file.

        Returns:
            image: returns the `NodeImage` object.
        """

        snapshot = self.create_snapshot(source)
        log.debug('Finished create snapshot: %r' % snapshot.id)

        log.info('Start to register the image '
                 'from the snapshot: %r' % snapshot.id)
        image = self.register_image(snapshot)
        log.debug('Finish registering the image with id: %r' % image.id)

        return image



# This file is part of fedimg.
# Copyright (C) 2014 Red Hat, Inc.
#
# fedimg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# fedimg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with fedimg; if not, see http://www.gnu.org/licenses,
# or write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Authors:  David Gay <dgay@redhat.com>
#

from setuptools import setup, find_packages

setup(
    name='fedimg',
    version='1.0.1',
    description='Automatically upload Fedora Cloud images to cloud providers',
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: GNU Affero General Public License"
        " v3 or later (AGPLv3+)",
    ],
    keywords='python Fedora cloud image uploader service',
    author='David Gay',
    author_email='dgay@redhat.com',
    url='https://github.com/fedora-infra/fedimg',
    license='AGPLv3+',
    include_package_data=True,
    zip_safe=False,
    install_requires=["fedmsg",
                      "apache-libcloud",
                      "paramiko"],
    tests_require=['nose',
                   'mock'],
    packages=find_packages(),
    entry_points="""
    [moksha.consumer]
    fedimgconsumer.prod = fedimg.consumers:FedimgConsumer
    fedimgconsumer.stg = fedimg.consumers:FedimgStagingConsumer
    fedimgconsumer.dev = fedimg.consumers:FedimgDevConsumer
    """,
)
